﻿# -*- coding: UTF-8 -*-
#brailleDisplayDrivers/handyTech.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright (C) 2008-2011 Felix Grützmacher <felix.gruetzmacher@handytech.de>, James Teh <jamie@jantrid.net>, Bram Duvigneau <bram@bramd.nl>

import _winreg
import comtypes.client
from comtypes import GUID
import braille
from logHandler import log
import speech
import inputCore
import os
from baseObject import ScriptableObject

COM_CLASS = "HtBrailleDriverServer.HtBrailleDriver"
HT_KEYS = {}

constants = None

class Sink(object):

	def __init__(self, server):
		self.server = server
		super(Sink, self).__init__()

	def sayString(self, text):
		speech.speakMessage(text)

	def onKeysPressed(self, keys_arg, routing_pos):
		# keys_arg is VARIANT. Indexing by 0 gives actual value.
		keys = keys_arg[0]
		if constants.KEY_ROUTING in keys:
			gesture = CommandGesture(keys, routing_pos -1)
		else:
			gesture = CommandGesture(keys)
		try:
			inputCore.manager.executeGesture(gesture)
		except inputCore.NoInputGestureAction:
			pass

	def onAtcEventAction(self, action, position, param1, param2):
		className = getClassNameFromBraillePos(position)
		if not className in (u'Edit', u'_WwG',):
			return # Not a window class in which we allow ATC assistant functions
		if action == constants.ATCACTION_DO_NOTHING:
			pass
		elif action == constants.ATCACTION_HIGH_SIGNAL:
			pass # COM server does this
		elif action == constants.ATCACTION_LINEEND_SIGNAL:
			pass # COM server does this
		elif action == constants.ATCACTION_LOW_SIGNAL:
			pass # COM server does this
		elif action == constants.ATCACTION_READ_ALOUD:
			pass # TODO
		elif action == constants.ATCACTION_ROUTE_CURSOR:
			pass # TODO
		elif action == constants.ATCACTION_SAY_CHARACTER:
			speech.speakMessage(getRawTextFromBrailleRange(param1-1, param1-1))
		elif action == constants.ATCACTION_SAY_LINE:
			speech.speakMessage(getRawTextFromBrailleRange(param1-1, param2-1))
		elif action == constants.ATCACTION_SAY_LINE_GO_NEXT_LINE:
			speech.speakMessage(getRawTextFromBrailleRange(param1-1, param2-1))
			braille.handler.buffer.scrollForward()
		elif action == constants.ATCACTION_SAY_WORD:
			speech.speakMessage(getRawTextFromBrailleRange(param1-1, param2-1))
		elif action == constants.ATCACTION_SHOW_NEXT_LINE:
			braille.handler.buffer.scrollForward()
		elif action == constants.ATCACTION_SHOW_PREV_LINE:
			braille.handler.buffer.scrollBack()

class BrailleDisplayDriver(braille.BrailleDisplayDriver, ScriptableObject):
	"""Handy Tech braille display driver.
	"""
	name = "handyTech"
	# Translators: Names of braille displays.
	description = _("Handy Tech braille displays")

	@classmethod
	def check(cls):
		try:
			GUID.from_progid(COM_CLASS)
			return True
		except comtypes.COMError:
			return False

	def __init__(self):
		global constants, HT_KEYS
		super(BrailleDisplayDriver, self).__init__()
		self._server = comtypes.client.CreateObject(COM_CLASS)
		import comtypes.gen.HTBRAILLEDRIVERSERVERLib as constants

		HT_KEYS = {}
		for key, constant in constants.__dict__.items():
			if key.startswith('KEY_'):
				HT_KEYS[constant] = key[4:].lower().replace('_', '')

		# Keep the connection object so it won't become garbage
		self._advise = comtypes.client.GetEvents(self._server, Sink(self._server), constants.IHtBrailleDriverSink)
		self._server.initialize()

	def terminate(self):
		super(BrailleDisplayDriver, self).terminate()
		self._server.terminate()

	def _get_numCells(self):
		return self._server.getCurrentTextLength()[0]

	def display(self, cells):
		self._server.displayText(cells)

	def script_showConfig(self, gesture):
		self._server.startConfigDialog(False)
	script_showConfig.__doc__ = _("Show the Handy Tech driver configuration window.")

	gestureMap = inputCore.GlobalGestureMap({
		"globalCommands.GlobalCommands": {
			"braille_scrollBack": ("br(handytech):left", "br(handytech):up"),
			"braille_previousLine": ("br(handytech):b4",),
			"braille_nextLine": ("br(handytech):b5",),
			"braille_scrollForward": ("br(handytech):right", "br(handytech):down"),
			"braille_routeTo": ("br(handytech):routing",),
			"kb:shift+tab": ("br(handytech):esc",),
			"kb:alt": ("br(handytech):b2+b4+b5",),
			"kb:escape": ("br(handytech):b4+b6",),
			"kb:tab": ("br(handytech):enter",),
			"kb:enter": ("br(handytech):esc+enter",),
			"kb:upArrow": ("br(handytech):leftSpace",),
			"kb:downArrow": ("br(handytech):rightSpace",),
			"showGui": ("br(handytech):b2+b4+b5+b6",),
		}
	})

	__gestures = {
		'br(handytech):b4+b8': 'showConfig',
	}

class CommandGesture(braille.BrailleDisplayGesture):

	source = BrailleDisplayDriver.name

	def __init__(self, keys, routing_pos=None):
		super(CommandGesture, self).__init__()
		self.keys = set(keys)

		self.keyNames = names = set()
		for key in self.keys:
			if key == constants.KEY_ROUTING:
				names.add("routing")
				self.routingIndex = routing_pos
			else:
				try:
					names.add(HT_KEYS[key])
				except KeyError:
					log.debugWarning("Unknown key %d" % key)

		self.id = "+".join(names)

def getRawTextFromBrailleRange(windowStartPos, windowEndPos):
	buffer = braille.handler.buffer
	bufferStartPos = buffer.windowStartPos + windowStartPos
	bufferEndPos = buffer.windowStartPos + windowEndPos
	if bufferEndPos >= buffer.windowEndPos: ## Beyond displayed text
		return u''
	startRegion, startPos = buffer.bufferPosToRegionPos(bufferStartPos)
	endRegion, endPos = buffer.bufferPosToRegionPos(bufferEndPos)
	if not startRegion is endRegion: # We don't fetch across regions
		return u''
	if endPos >= len(startRegion.brailleToRawPos):
		return u''
	rawStartPos = startRegion.brailleToRawPos[startPos]
	rawEndPos = startRegion.brailleToRawPos[endPos]
	return startRegion.rawText[rawStartPos:rawEndPos+1]

def getClassNameFromBraillePos(windowPos):
	buffer = braille.handler.buffer
	bufferPos = buffer.windowStartPos + windowPos
	if bufferPos >= buffer.windowEndPos:
		return u'Edit'
	region, pos = buffer.bufferPosToRegionPos(bufferPos)
	if region is None:
		return u'No region.'
	if not isinstance(region, braille.NVDAObjectRegion) and not isinstance(region, braille.TextInfoRegion):
		return u'Not tied to an object.'
	import NVDAObjects.window
	if not isinstance(region.obj, NVDAObjects.window.Window):
		return u'Not a window.'
	return region.obj.windowClassName
